-- GLOBALS: unpack, VoidStorageFrame

--[[//////////////////////////////////////////////////

    GLO-LO

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]

    local _
    local select = select
    local table = table
    local tostring = tostring
    local next = next
    local tonumber = tonumber
    local pairs = pairs
    local ipairs = ipairs
    local math = math
    local string = string
    local type = type
    local strsplit = strsplit
    local strmatch = strmatch
    local strfind = strfind

    local GetTime = GetTime
    local GetItemInfo = GetItemInfo
    local hooksecurefunc = hooksecurefunc
    local GetAddOnInfo = GetAddOnInfo
    local GetRealmName = GetRealmName
    local UnitName = UnitName
    local UnitFactionGroup = UnitFactionGroup
    local GameTooltip = GameTooltip
    local GetCurrentGuildBankTab = GetCurrentGuildBankTab
    local PlaySoundFile = PlaySoundFile
    local GetVoidItemHyperlinkString = GetVoidItemHyperlinkString
    local GetContainerItemInfo = GetContainerItemInfo
    local GetInventoryItemLink = GetInventoryItemLink
    local GetItemQualityColor = GetItemQualityColor
    local GetContainerNumSlots = GetContainerNumSlots
    local GetInventorySlotInfo = GetInventorySlotInfo
    local GetContainerItemLink = GetContainerItemLink
    local ContainerIDToInventoryID = ContainerIDToInventoryID
    local BankButtonIDToInvSlotID = BankButtonIDToInvSlotID
    local IsArtifactPowerItem = IsArtifactPowerItem
    local GetItemCount = GetItemCount
    local GetItemIcon = GetItemIcon
    local GetCoinIcon = GetCoinIcon
    local FauxScrollFrame_GetOffset = FauxScrollFrame_GetOffset
    local FauxScrollFrame_Update = FauxScrollFrame_Update
    local StaticPopup_Show = StaticPopup_Show
    local IsShiftKeyDown = IsShiftKeyDown
    local CloseDropDownMenus = CloseDropDownMenus
    local IsVoidStorageReady = IsVoidStorageReady
    local GetVoidItemInfo = GetVoidItemInfo
    local GetGuildBankItemInfo = GetGuildBankItemInfo
    local GetGuildBankItemLink = GetGuildBankItemLink
    local GetNumAuctionItems = GetNumAuctionItems
    local GetAuctionItemInfo = GetAuctionItemInfo
    local GetAuctionItemLink = GetAuctionItemLink
    local GetAuctionItemTimeLeft = GetAuctionItemTimeLeft
    local GetInboxNumItems = GetInboxNumItems
    local GetInboxHeaderInfo = GetInboxHeaderInfo
    local GetInboxItem = GetInboxItem
    local GetInboxItemLink = GetInboxItemLink
    local GetNumGuildBankTabs = GetNumGuildBankTabs
    local IsInGuild = IsInGuild
    local GetGuildInfo = GetGuildInfo
    local GetGuildBankTabInfo = GetGuildBankTabInfo

    local LibStub = LibStub

    local B, L, T = unpack(select(2, ...))
    local Icon = LibStub("LibDBIcon-1.0")

    T.TOOLTIPLABEL = {
        ['bank'] = L[10013],
        ['reagent'] = L[10014],
        ['void'] = L[10016],
        ['guild'] = L[10017],
        ['player'] = L[10015],
        ['equipped'] = L[10012],
        ['auctions'] = L[9001],
        ['mail'] = L[9000]
    }

--[[//////////////////////////////////////////////////

    MAIN FUNCTIONS

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]
    
    function B.CloseSpecialWindows()
        T.F.mainframe.lastbank = nil
    end

    function B.Initialize()
        local enabled, addons

        hooksecurefunc("ContainerFrameItemButton_OnModifiedClick", B.ContainerFrameItemButton_OnModifiedClick)
        hooksecurefunc("ChatFrame_OnHyperlinkShow", B.ChatFrame_OnHyperlinkShow)
        hooksecurefunc("ChatEdit_InsertLink", B.InsertLink)
        hooksecurefunc("SetItemButtonQuality", B.ButtonListener) -- Magic workaround multiple event triggering
        hooksecurefunc("CloseSpecialWindows", B.CloseSpecialWindows)

        addons = {"Adibags", "ElvUI"} -- Addons that modify frame names
        for x = 1, #addons do
            _, _, _, enabled = GetAddOnInfo(addons[x])
            if enabled then
                T.BW.forceBankScan = true
                break
            end
        end

        local LibExtraTip = LibStub("LibExtraTip-1")
        LibExtraTip:AddCallback({type = "item", callback = B.TooltipListener})
        LibExtraTip:RegisterTooltip(GameTooltip)

        local player = UnitName('player')
        local server = GetRealmName()
        local faction = UnitFactionGroup('player')

        B.Load(player, server, faction)

        local BanksWatchLDB = LibStub("LibDataBroker-1.1"):NewDataObject("BanksWatchMinimap", {
            type = "data source",
            text = "Bank's Watch",
            icon = T.icon,
            iconR = T.CLASS.r,
            iconG = T.CLASS.g,
            iconB = T.CLASS.b,
            OnClick = function(self, button)
                if button == "LeftButton" then
                    B.Main("toggle")
                elseif button == "RightButton" then
                    if not InterfaceOptionsFrame:IsShown() then
                        InterfaceOptionsFrame_Show() -- Open options
                    end
                    InterfaceOptionsFrame_OpenToCategory(B.C['panel']) -- Bank's Watch
                end
            end,
            OnTooltipShow = function()
                GameTooltip:SetText("Bank's Watch")
                GameTooltip:AddLine(" ")
                GameTooltip:AddLine("|cffffffff"..L[10066])
                GameTooltip:AddLine("|cffffffff"..L[10067])
            end
        })

        Icon:Register("BanksWatchMinimap", BanksWatchLDB, BWsavedtable['minimapicon'])
        if BWsavedtable['minimapicon']['hide'] then
            Icon:Hide("BanksWatchMinimap")
        else
            Icon:Show("BanksWatchMinimap")
        end
    end

    function B.Load(player, server, faction)

        if (player == nil) or (player == L[9002]) or (player == L[9003]) then
            return
        end

        if BWsavedtable and BWsavedtable[server] and BWsavedtable[server][player.."."..string.sub(faction, 1, 1)] then
            B.load(server, player, faction)
        else
            B.print(T.MSGC.EVENT.."New character detected")
            B.template(server, player, faction)
        end
    end

    function B.Main(input)

        if input then input = string.lower(input) end
        local input, arg1, val1, val2 = strsplit(" ", input)

        if input == "toggle" then
            if T.F.toggleframe:IsVisible() then
                T.F.toggleframe:Hide()
            else
                T.F.toggleframe:Show()
            end

        elseif input == "minimap" then
            BWsavedtable['minimapicon']['hide'] = not BWsavedtable['minimapicon']['hide']
            if BWsavedtable['minimapicon']['hide'] then
                Icon:Hide("BanksWatchMinimap")
            else
                Icon:Show("BanksWatchMinimap")
            end

        elseif input == "debug" then
            BWdebug = not BWdebug
            B.print(T.MSGC.INFO.."Debug:|r "..T.MSGC.BOOL..tostring(BWdebug), 1)

        elseif input == "item" then
            T.F.itemframe:Show()

        elseif input == "money" then

            if arg1 then
                if arg1 == "scale" then
                -- verification to make sure nothing silly is happening
                    if not tonumber(val1) then B.print(T.MSGC.ERROR..L[10055], 1); return end

                    val1 = math.max(val1, 75)
                    val1 = math.min(val1, 200)

                    T.CONFIG['scalemoneytracker'] = tonumber(val1)
                    B.print(T.MSGC.INFO.."Scaling:|r "..tostring(T.CONFIG['scalemoneytracker'] / 100), 1)
                    B.M.Frame.Scale(val1)
                    B.Main("money test")

                elseif arg1 == "test" then
                    B.M.Trigger(tonumber(val1 or 100000))

                elseif arg1 == "toggle" then
                    T.CONFIG['enablemoneytracker'] = not T.CONFIG['enablemoneytracker']
                    B.print(T.MSGC.INFO.."Small money tracker:|r "..tostring(T.CONFIG['enablemoneytracker']), 1)
                end
            end
        end
    end

    function B.ButtonListener(button, quality, link)
        -- Workaround to using events which may trigger more than once.
        --B.print(button:GetName())

        if T.BW.loaded then
            if B.IsGuildBank(button) then
                B.GuildSlotScan(GetCurrentGuildBankTab(), button:GetID(), button, link)
            
            elseif B.IsPlayerBag(button) then
                B.PlayerBagSlotScan(button:GetParent():GetID(), button:GetID())

            elseif B.IsBankBag(button) then
                B.BankBagSlotScan(button:GetParent():GetID(), button:GetID())
            
            elseif B.IsReagentBank(button) then
                B.ReagentSlotScan(button:GetID())

            elseif B.IsGenericBank(button) then
                B.GenericBankSlotScan(button:GetID())

            elseif B.IsVoidBank(button) then
                B.VoidSlotScan(tonumber(VoidStorageFrame.page), button:GetID(), button)
            end
        end
    end

    local newstring
    local order = {"player", "equipped", "bank", "reagent", "void", "guild", "mail", "auctions"}
    function B.TooltipListener(tooltip, itemlink, quantity)
        -- Minimal approach
        newstring = newstring or B.newtable()

        local result, itemid, bank, bag, p, k, v, total, total_server, guild, empty_line, server_line, class_c
        local player = UnitName('player')
        local faction = string.sub(UnitFactionGroup('player'), 1, 1)
        local server = GetRealmName()
        local alt, altfaction, altserver

        local r, g, b = T.TOOLTIP.r, T.TOOLTIP.g, T.TOOLTIP.b --0.74, 0.46, 1.00, nice purple...
        local contrast = B.hex(T.TOOLTIPAC.r, T.TOOLTIPAC.g, T.TOOLTIPAC.b)

        local total_players = 0
        local found_alts = 0
        local ignored = {
            [6948] = true, -- Hearthstone
            [110560] = true, -- Garrison Hearthstone
            [140192] = true -- Dalaran Hearthstone
        }

        if itemlink then
            itemid = tonumber(B.GetItemIdFromLink(itemlink))
            if ignored[itemid] then return end

            table.wipe(newstring)
            total = 0
            total_server = 0
            empty_line = false

            -- Main player
            for x = 1, #order do
                bank = order[x]
                bag = T.BW.tooltipcache[server][player.."."..faction][bank]

                if bag then
                    result = bag[itemid]
                    if result then
                        newstring[#newstring + 1] = contrast..T.TOOLTIPLABEL[bank].."|r: |cffffffff"..result.."|r"
                        if bank == "guild" or bank == "mail" then
                            total_server = total_server + result
                        elseif bank ~= "auctions" then
                            total = total + result
                        end
                        total_players = total_players + result
                    end
                end
            end

            if next(newstring) then
                tooltip:AddLine(" ")
                found_alts = found_alts + 1
                empty_line = true
                server_line = total_server > 0 and "/|cffffd100"..total_server + total.."|r " or ""
                tooltip:AddDoubleLine("|c"..T.RAID_CLASS_COLORS[T['class']].colorStr..player.."|r"..contrast.."|r:|cffffffff "..total.." |r"..server_line..contrast.."|r", table.concat(newstring, ", "), r, g, b, r, g, b)
            end

            -- Alts from same server
            table.wipe(newstring)
            for p, bank in pairs(T.BW.tooltipcache[server]) do
                if p ~= player.."."..faction then
                    alt, altfaction = p:match"([^.]*).(.*)"
                    if altfaction == faction or BWsavedtable['enablesearchallfactions'] then
                        for x = 1, #order do
                            bank = order[x]
                            bag = T.BW.tooltipcache[server][p][bank]

                            if bag then
                                result = bag[itemid]
                                if result then
                                    newstring[#newstring + 1] = contrast..T.TOOLTIPLABEL[bank].."|r: |cffffffff"..result.."|r"
                                    total_players = total_players + result
                                end
                            end
                        end
                        if next(newstring) then
                            if not empty_line then tooltip:AddLine(" "); empty_line = true end
                            found_alts = found_alts + 1
                            class_c = BWsavedtable[server][p]['player']['class'] and T.RAID_CLASS_COLORS[BWsavedtable[server][p]['player']['class']].colorStr or "ffffffff"
                            tooltip:AddDoubleLine("|c"..class_c..alt.."|r"..(BWsavedtable['enablesearchallfactions'] and " "..altfaction or ""), table.concat(newstring, ", "), r, g, b, r, g, b)
                        end
                        table.wipe(newstring)
                    end
                end
            end

            -- Alts from other server
            if BWsavedtable['enablesearchallservers'] then
                table.wipe(newstring)
                for p, bank in pairs(T.BW.tooltipcache) do
                    if p ~= server then
                        empty_line = false
                        for k, v in pairs(bank) do
                            alt, altfaction = k:match"([^.]*).(.*)"
                            if BWsavedtable['enablesearchallfactions'] or altfaction == faction then
                                for x = 1, #order do
                                    bank = order[x]
                                    bag = T.BW.tooltipcache[p][k][bank]

                                    if bag then
                                        result = bag[itemid]
                                        if result then
                                            newstring[#newstring + 1] = contrast..T.TOOLTIPLABEL[bank].."|r: |cffffffff"..result.."|r"
                                            total_players = total_players + result
                                        end
                                    end
                                end
                                if next(newstring) then
                                    if not empty_line then tooltip:AddLine(" "); tooltip:AddLine(p); empty_line = true end
                                    found_alts = found_alts + 1
                                    class_c = BWsavedtable[p][k]['player']['class'] and T.RAID_CLASS_COLORS[BWsavedtable[p][k]['player']['class']].colorStr or "ffffffff"
                                    tooltip:AddDoubleLine("|c"..class_c..alt.."|r"..(BWsavedtable['enablesearchallfactions'] and " "..altfaction or ""), table.concat(newstring, ", "), r, g, b, r, g, b)
                                end
                                table.wipe(newstring)
                            end
                        end
                    end
                end
            end

            if found_alts > 1 then
                tooltip:AddLine(" ")
                tooltip:AddDoubleLine("", "|cffffd100Total:|r |cffffffff"..total_players.."|r", r, g, b, r, g, b)
            end
        end
    end

    function B.ContainerFrameItemButton_OnModifiedClick(self)

        if IsShiftKeyDown() then
            local bag, item = self:GetParent():GetID(), self:GetID()
            local link = GetContainerItemLink(bag, item)

            if T.F.mainframe.searchbar:HasFocus() and T.F.mainframe.searchbar:GetText() ~= link then

                T.F.mainframe.searchbar:SetText(GetItemInfo(link))
            end
        end
    end

    function B.ChatFrame_OnHyperlinkShow(self, link, text, button)

        if IsShiftKeyDown() and link then
            if string.find(link, 'achievement:*') or string.find(link,'spell:*') then
                return false
            end

            if T.F.mainframe.searchbar:HasFocus() and T.F.mainframe.searchbar:GetText() ~= text then
                T.F.mainframe.searchbar:SetText(GetItemInfo(link))
            end
        end
    end

    function B.InsertLink(text)

        if T.F.mainframe.searchbar:HasFocus() and T.F.mainframe.searchbar:GetText() ~= text then
            T.F.mainframe.searchbar:SetText(string.find(text, "^|c%x+|H(.+)|h%[.*%]"))
        end
    end

    function B.dropDownMenuItem_OnClick(self)
        local server, player, faction
        local value = self.value

        if not value then return end

        if value == "delete" then
            BWsavedtable[T.ACTIVE_TABLE.server][T.ACTIVE_TABLE.player.."."..string.sub(T.ACTIVE_TABLE.faction, 1, 1)] = nil

            -- Verify if server still needs to exists.
            if not next(BWsavedtable[T.ACTIVE_TABLE.server]) then BWsavedtable[T.ACTIVE_TABLE.server] = nil end

            local player = UnitName('player')
            local server = GetRealmName()
            local faction = UnitFactionGroup('player')
            B.Load(player, server, faction)

            if B.IsCurrentPlayer() then
                B.PlayerBagsScan()
                B.PlayerEquipmentScan()
            end

            B.UpdateBanks()
            -- Cache items here
            B.CacheMainItems()
            B.CacheAltItems()

            T.F.mainframe.character.auctions:Hide()
            T.F.mainframe.character.mail:Hide()

        elseif value == "searchallservers" then
            BWsavedtable['enablesearchallservers'] = not BWsavedtable['enablesearchallservers']
            return

        elseif value == "searchallfactions" then
            BWsavedtable['enablesearchallfactions'] = not BWsavedtable['enablesearchallfactions']
            return

        elseif value == "movemoneytracker" then
            T.F.moneyframe.anchor:Show()

        elseif value == "enablesoundsmoneytracker" then
            T.CONFIG['enablesoundsmoneytracker'] = not T.CONFIG['enablesoundsmoneytracker']
            return
        elseif value == "enablemoneytracker" then
            T.CONFIG['enablemoneytracker'] = not T.CONFIG['enablemoneytracker']
            return
        elseif value == "minimummoneytracker" then StaticPopup_Show("BWsetminimum")
        elseif value == "treasuremoneytracker" then StaticPopup_Show("BWsettreasure")
        elseif value == "coinssound" then
            T.CONFIG['moneysound'] = 1
            PlaySoundFile(T.SOUNDS[T.CONFIG['moneysound']], "Master")
            return

        elseif value == "retrocoinssound" then
            T.CONFIG['moneysound'] = 2
            PlaySoundFile(T.SOUNDS[T.CONFIG['moneysound']], "Master")
            return

        elseif value == "treasure1sound" then
            T.CONFIG['treasuresound'] = 3
            PlaySoundFile(T.SOUNDS[T.CONFIG['treasuresound']], "Master")
            return

        elseif value == "treasure2sound" then
            T.CONFIG['treasuresound'] = 4
            PlaySoundFile(T.SOUNDS[T.CONFIG['treasuresound']], "Master")
            return
        else
            server, player, faction = value:match"([^.]*).([^.]*).(.*)"
            faction = faction == "H" and "Horde" or faction == "A" and "Alliance" or faction == "N" and "Neutral"
            B.load(server, player, faction)
        end

        CloseDropDownMenus()
    end

    function B.UpdateBanks()

        B.BankUpdate()
        B.ReagentUpdate()
        B.VoidUpdate()
        B.GuildUpdate()
        B.PlayerUpdate()
        B.PlayerEquipmentUpdate()

        local text = T.F.mainframe.searchbar:GetText()
        if text and text ~= "" then
            T.F.mainframe.searchbar:SetText("")
            T.F.mainframe.searchbar:SetText(text)
        else
            B.Search(false)
        end

        B.ScrollbarUpdateAuctions()
        B.ScrollbarUpdateMail()
    end

--[[//////////////////////////////////////////////////

    SCAN FUNCTIONS (OLDER)

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]
    
    function B.BankScan()

        if not T.ISBANKOPEN then return end -- Circumvent early BAG_UPDATE
        local x, counter, count, link, bag, slot, slots
        local bank = T['bank']

        if not bank[1] then bank[1] = B.newtable() end
        for x = 1, T.NUM_BANKGENERIC_SLOTS do
            _, count, _, _, _, _, link = GetContainerItemInfo(T.BANK_CONTAINER, x)

            if link then
                bank[1][x] = bank[1][x] or B.newtable()
                bank[1][x]['count'] = count
                bank[1][x]['link'] = link
            else
                bank[1][x] = B.deltable(bank[1][x])
            end
        end

        for bag = T.ITEM_INVENTORY_BANK_BAG_OFFSET + 1, T.ITEM_INVENTORY_BANK_BAG_OFFSET + T.NUM_BANKBAGSLOTS do
            if not bank[bag] then bank[bag] = B.newtable() end
            
            local inventoryID = BankButtonIDToInvSlotID(bag, 1) - 4
            local baglink = GetInventoryItemLink("player", inventoryID)

            slots = GetContainerNumSlots(bag)
            bank[bag]['slots'] = slots
            bank[bag]['link'] = baglink

            for x = 1, slots do
                _, count, _, _, _, _, link = GetContainerItemInfo(bag, x)

                if link then
                    bank[bag][x] = bank[bag][x] or B.newtable()
                    bank[bag][x]['link'] = link
                    bank[bag][x]['count'] = count > 1 and count or nil
                else
                    bank[bag][x] = B.deltable(bank[bag][x])
                end
            end
        end

        if B.IsCurrentPlayer() then B.BankUpdate() end
        B.CacheMainItems('bank')
    end

    function B.ReagentScan()

        local x, page, itemID
        local reagent = T['reagent']

        for page = 1, T.REAGENT_PAGES do
            if not reagent[page] then reagent[page] = B.newtable() end

            for x = 1, T.NUM_REAGENT_SLOTS do

                local _, count, _, _, _, _, link = GetContainerItemInfo(T.REAGENTBANK_CONTAINER, x)

                if link then
                    reagent[page][x] = reagent[page][x] or B.newtable()
                    reagent[page][x]['link'] = link
                    reagent[page][x]['count'] = count > 1 and count or nil
                else
                    reagent[page][x] = B.deltable(reagent[page][x])
                end
            end
        end
    end

    function B.VoidScan()

        if not IsVoidStorageReady() then return end
        local x, page, itemID
        local void = T['void']

        for page = 1, T.VOID_PAGES do
            if not void[page] then void[page] = B.newtable() end

            for x = 1, T.NUM_VOID_SLOTS do
                itemID = GetVoidItemInfo(page, x)

                if itemID then
                    void[page][x] = void[page][x] or B.newtable()
                    void[page][x]['link'] = page == 1 and GetVoidItemHyperlinkString(x) or GetVoidItemHyperlinkString(T.NUM_VOID_SLOTS + x)
                else
                    void[page][x] = B.deltable(void[page][x])
                end
            end
        end
    end

    function B.GuildScan()

        local x, i, id, count, item, button
        local guild = T['guild']
        local page = GetCurrentGuildBankTab()

        if guild[page].access then
            for x = 1, T.MAX_GUILDBANK_SLOTS_PER_TAB do

                _, count = GetGuildBankItemInfo(page, x)
                if count > 0 then
                    guild[page][x] = guild[page][x] or B.newtable()
                    guild[page][x]['link'] = GetGuildBankItemLink(page, x)
                    guild[page][x]['count'] = count > 1 and count or nil
                else
                    guild[page][x] = B.deltable(guild[page][x])
                end
            end
        end
    end

--[[//////////////////////////////////////////////////

    SCAN FUNCTIONS

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]

    function B.Scan(slot, link, count)

        if link then
            slot = slot or B.newtable()
            slot['link'] = link
            slot['count'] = count and count > 1 and count or nil
        else
            slot = B.deltable(slot)
        end

        return slot
    end

    function B.BankBagScan(bag)

        local x, slots, count, link
        local bank = T['bank']

        B.RefreshBankBag(bank, bag)

        for x = 1, bank[bag]['slots'] do
            _, count, _, _, _, _, link = GetContainerItemInfo(bag, x)
            bank[bag][x] = B.Scan(bank[bag][x], link, count)
        end

        if B.IsCurrentPlayer() then B.BankUpdate() end
        B.CacheMainItems('bank')
    end

    function B.GenericBankSlotScan(slot)

        local bank = T['bank']
        local _, count, _, _, _, _, link = GetContainerItemInfo(T.BANK_CONTAINER, slot)

        if not B.CompareSlot(bank[1][slot], link, count) then

            bank[1][slot] = B.Scan(bank[1][slot], link, count)
            if B.IsCurrentPlayer() then B.GenericBankSlotUpdate(slot) end
            B.CacheMainItems('bank')
        end
    end

    function B.BankBagSlotScan(bag, slot)

        local bank = T['bank']
        local _, count, _, _, _, _, link = GetContainerItemInfo(bag, slot)

        B.RefreshBankBag(bank, bag)

        if not B.CompareSlot(bank[bag][slot], link, count) then

            bank[bag][slot] = B.Scan(bank[bag][slot], link, count)
            if B.IsCurrentPlayer() then B.BankSlotUpdate(B.GetSlotID(bag, slot), bag, slot) end
            B.CacheMainItems('bank')
        end
    end

    function B.ReagentSlotScan(slot)

        local reagent = T['reagent']
        local _, count, _, _, _, _, link = GetContainerItemInfo(T.REAGENTBANK_CONTAINER, slot)

        if not reagent[1] then reagent[1] = B.newtable() end

        if not B.CompareSlot(reagent[1][slot], link, count) then

            reagent[1][slot] = B.Scan(reagent[1][slot], link, count)
            if B.IsCurrentPlayer() then B.ReagentSlotUpdate(slot) end
            B.CacheMainItems('reagent')
        end
    end

    function B.VoidSlotScan(page, slot, button)

        local void = T['void']
        local link = nil
        if button.hasItem then link = page == 1 and GetVoidItemHyperlinkString(slot) or GetVoidItemHyperlinkString(T.NUM_VOID_SLOTS + slot) end

        if not void[page] then void[page] = B.newtable() end

        if not B.CompareSlot(void[page][slot], link) then

            void[page][slot] = B.Scan(void[page][slot], link)
            if B.IsCurrentPlayer() then B.VoidSlotUpdate(page, slot) end
            B.CacheMainItems('void')
        end
    end

    function B.GuildSlotScan(page, slot, button, link)

        local guild = T['guild']

        if button.count > 0 then
            if strfind(link, "item:82800:") then -- Caged pet
                link = B.GetGuildPetLink(page, slot)
            end
        end

        if not B.CompareSlot(guild[page][slot], link, button.count) then

            guild[page][slot] = B.Scan(guild[page][slot], link, button.count)
            if B.IsCurrentPlayer() then B.GuildSlotUpdate(page, slot) end
            B.CacheMainItems('guild')
        end
    end

    function B.PlayerBagsScan()

        local x, counter, count, link, bag, slot, slots
        local player = T['player']

        for bag = T.BACKPACK_CONTAINER, T.NUM_BAG_SLOTS do
            if not player[bag] then player[bag] = B.newtable() end

            slots = GetContainerNumSlots(bag)
            player[bag]['slots'] = slots

            if bag > 0 then
                local inventoryID = ContainerIDToInventoryID(bag)
                local baglink = GetInventoryItemLink("player", inventoryID)
                player[bag]['link'] = baglink
            end

            for x = 1, slots do
                _, count, _, _, _, _, link = GetContainerItemInfo(bag, x)

                player[bag][x] = B.Scan(player[bag][x], link, count)
            end
        end

        if B.IsCurrentPlayer() then B.PlayerUpdate() end
        B.CacheMainItems('player')
    end

    function B.PlayerBagScan(bag)

        local x, slots, count, link
        local player = T['player']

        if not player[bag] then player[bag] = B.newtable() end
            
        slots = GetContainerNumSlots(bag)
        player[bag]['slots'] = slots

        if bag > 0 then
            local inventoryID = ContainerIDToInventoryID(bag)
            local baglink = GetInventoryItemLink("player", inventoryID)
            player[bag]['link'] = baglink
        end

        for x = 1, slots do
            _, count, _, _, _, _, link = GetContainerItemInfo(bag, x)
            player[bag][x] = B.Scan(player[bag][x], link, count)
        end

        if B.IsCurrentPlayer() then B.PlayerUpdate() end
        B.CacheMainItems('player')
    end

    function B.PlayerBagSlotScan(bag, slot)

        local player = T['player']
        local _, count, _, _, _, _, link = GetContainerItemInfo(bag, slot)

        if not player[bag] then
            player[bag] = B.newtable()
            local slots = GetContainerNumSlots(bag)
            player[bag]['slots'] = slots

            if bag > 0 then
                local inventoryID = ContainerIDToInventoryID(bag)
                local baglink = GetInventoryItemLink("player", inventoryID)
                player[bag]['link'] = baglink
            end
        end

        if not B.CompareSlot(player[bag][slot], link, count) then

            player[bag][slot] = B.Scan(player[bag][slot], link, count)
            if B.IsCurrentPlayer() then B.PlayerSlotUpdate(bag, slot) end
            B.CacheMainItems('player')
        end
    end

    function B.PlayerEquipmentScan()

        local inventory, slotId, link
        local player = T['player']

        if not player[-1] then player[-1] = B.newtable() end

        -- Tabard slot is slotId 19?
        for x = 1, #T.INVSLOTS do

            slotId = GetInventorySlotInfo(T.INVSLOTS[x])
            link = GetInventoryItemLink("player", slotId)

            if slotId == 19 then slotId = 18 end

            player[-1][slotId] = B.Scan(player[-1][slotId], link)
        end

        B.CacheMainItems('equipped')
    end

    function B.PlayerEquipmentSlotScan(slot)

        local player = T['player']
        local link = GetInventoryItemLink("player", slot)

        if slot == 19 then slot = 18 end

        if not B.CompareSlot(player[-1][slot], link, nil) then

            player[-1][slot] = B.Scan(player[-1][slot], link)
            if B.IsCurrentPlayer() then B.PlayerEquipmentSlotUpdate(slot) end
            B.CacheMainItems('equipped')
        end
    end

    function B.BankOfflineScan(bag)
        -- Item count is always updated. Check the difference for the bag.
        local slot, count, total
        local bank = T['bank']

        total = 0
        for x = 1, bank[bag]['slots'] do
            slot = bank[bag][x]
            if slot and next(slot) then
                count = (GetItemCount(slot['link'], true) or 0) - (GetItemInfo(slot['link']) or 0)
                B.print(slot['link'])
                B.print(count)
                --[[if count then -- item just changed
                    B.print(slot['link'])
                    B.print(count)
                    B.print(slot['count'])
                end]]
            end
        end
    end

    function B.AuctionScan()
        local x, count, status, index, price, bid, bidAmount
        local auctions = T['auctions']
        local _, totalAuctions = GetNumAuctionItems('owner')

        table.wipe(auctions[1])

        for x = 1, totalAuctions do
            _, _, count, _, _, _, _, bid, _, price, bidAmount, _, _, _, _, status = GetAuctionItemInfo('owner', x)
            if status == 0 then
                index = #auctions[1] + 1
                auctions[1][index] = B.Scan(auctions[1][index], GetAuctionItemLink('owner', x), count)
                auctions[1][index]['time'] = GetTime() + T.TIMEINDEX[GetAuctionItemTimeLeft('owner', x)] -- when auction will actually expire
                auctions[1][index]['price'] = price > 0 and price or bidAmount > 0 and bidAmount or bid > 0 and bid
            end
        end
    end

    function B.MailScan()

        local x, a, index
        local mail = T['mail']
        local totalMoney, totalCOD = 0, 0
        local sender, money, CODAmount, daysLeft, itemCount
        local name, count

        table.wipe(mail[1])

        -- Can only load 50 mail
        for x = 1, GetInboxNumItems() do
            _, _, sender, _, money, CODAmount, daysLeft, itemCount = GetInboxHeaderInfo(x)

            if money > 0 then totalMoney = totalMoney + money end
            if CODAmount > 0 then totalCOD = totalCOD + CODAmount end

            if itemCount then
                index = #mail[1] + 1
                for a = 1, itemCount do
                    name, _, _, count = GetInboxItem(x, a)
                    mail[1][index] = B.Scan(mail[1][index], GetInboxItemLink(x, a), count)
                    if mail[1][index] then
                        mail[1][index]['time'] = GetTime() + (daysLeft*24*60*60)
                        mail[1][index]['money'] = money > 0 and money or CODAmount > 0 and -CODAmount or nil
                        mail[1][index]['sender'] = sender
                    end
                end
            end
        end

        mail['money'] = totalMoney
        mail['cod'] = totalCOD
    end

--[[//////////////////////////////////////////////////

    UPDATE FUNCTIONS

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]
    
    function B.ScrollbarUpdate()
        local r, g, b, fr, fg, fb, idx, link, hexcolor, buttonnum, item
        local player, server, location
        local quality, quest, icon, name, ilvl, faction
        local total_items = #T.BW.items
        local currentscrollbarvalue = FauxScrollFrame_GetOffset(T.F.mainframe.search.scroll)
        local rows = B.RowsToFit(T.F.mainframe.search.list, 32)

        FauxScrollFrame_Update(T.F.mainframe.search.scroll, total_items, rows, 32)
        
        for x = 1, rows do
            -- Clear buttons
            if T.F.mainframe.search.list.buttons[x] then 
                B.ResetButton(T.F.mainframe.search.list.buttons[x])
                --T.F.mainframe.search.list.buttons[x]:SetBackdropColor(T.THEME.r, T.THEME.g, T.THEME.b, 0.7)
                T.F.mainframe.search.list.buttons[x]:SetBackdropBorderColor(T.THEME.r, T.THEME.g, T.THEME.b, 1)
                --T.F.mainframe.search.list.buttons[x].rightstring.bg:SetVertexColor(T.THEME.r, T.THEME.g, T.THEME.b, 1)
                T.F.mainframe.search.list.buttons[x].altstring.faction:Hide()
                T.F.mainframe.search.list.buttons[x].leftstring:SetText("")
                T.F.mainframe.search.list.buttons[x]:Hide()
                T.F.mainframe.search.enddot:Hide()
            end
        end

        for x = 1, rows do
            -- Get the appropriate item, which will be x + value
            if (T.F.mainframe.layout == 2 or T.F.mainframe.layout == 4) and total_items > 0 and total_items < B.RowsToFit(T.F.mainframe.search.list) then
                local difference = B.RowsToFit(T.F.mainframe.search.list) - total_items
                buttonnum = x + difference
            else
                buttonnum = x
            end

            if T.F.mainframe.search.list.buttons[buttonnum] then

                idx = x + currentscrollbarvalue

                if T.BW.items[idx] then
                    link = T.BW.items[idx].link
                    item = B.GetItem(link)
                    name, quality, quest, icon, ilvl = item[1], item[3], (item[6] == L[9006]), item[10], item[4] or nil

                    if quality then
                        r, g, b, hexcolor = GetItemQualityColor(quality)

                        T.F.mainframe.search.list.buttons[buttonnum].quest = quest
                        T.F.mainframe.search.list.buttons[buttonnum].quality = quality
                        T.F.mainframe.search.list.buttons[buttonnum].icon:SetTexture(icon)
                        T.F.mainframe.search.list.buttons[buttonnum].leftstring:SetText("|c"..hexcolor..name)

                        if T.BW.items[idx].count > 1 then
                            T.F.mainframe.search.list.buttons[buttonnum].text:SetText(T.BW.items[idx].count)
                        elseif (item[6] == L[9004] or item[6] == L[9005]) and quality > 1  then
                            ilvl = B.DetectILVL(link) or ilvl
                            T.F.mainframe.search.list.buttons[buttonnum].ilvl:SetText(ilvl)
                            T.F.mainframe.search.list.buttons[buttonnum].ilvl:SetTextColor(r, g, b, 1)
                        end

                        if T.BW.items[idx].loc == "alt" then
                            server, player, faction = T.BW.items[idx].num:match"([^.]*).([^.]*).(.*)"
                            location = player

                            if faction ~= string.sub(T.ACTIVE_TABLE.faction, 1, 1) then
                                if T.ACTIVE_TABLE.faction == "Horde" then
                                    fr, fg, fb = 0.11, 0.44, 0.86
                                else -- Alliance
                                    fr, fg, fb = 0.74, 0.06, 0.19
                                end
                                T.F.mainframe.search.list.buttons[buttonnum].altstring.faction:SetVertexColor(fr, fg, fb, 1)
                                T.F.mainframe.search.list.buttons[buttonnum].altstring.faction:Show()
                            end

                            T.F.mainframe.search.list.buttons[buttonnum].altstring:SetText("|c"..(BWsavedtable[server][player.."."..string.sub(faction, 1, 1)]['player']['class'] and T.RAID_CLASS_COLORS[BWsavedtable[server][player.."."..string.sub(faction, 1, 1)]['player']['class']].colorStr or "ffffffff")..location)
                            T.F.mainframe.search.list.buttons[buttonnum].altstring.server:SetText(server)
                            T.F.mainframe.search.list.buttons[buttonnum].altstring:Show()
                            T.F.mainframe.search.list.buttons[buttonnum].altstring.server:Show()
                            T.F.mainframe.search.list.buttons[buttonnum].rightstring:Hide()
                        else
                            location = T.BW.items[idx].num == 1 and T.LABEL[T.BW.items[idx].loc] or T.LABEL[T.BW.items[idx].loc].." "..T.BW.items[idx].num
                            T.F.mainframe.search.list.buttons[buttonnum].rightstring:SetText(location)
                            T.F.mainframe.search.list.buttons[buttonnum].rightstring:Show()
                            T.F.mainframe.search.list.buttons[buttonnum].altstring:Hide()
                            T.F.mainframe.search.list.buttons[buttonnum].altstring.server:Hide()
                        end
                        T.F.mainframe.search.list.buttons[buttonnum].location = {['loc'] = T.BW.items[idx].loc, ['num'] = T.BW.items[idx].num}
                        T.F.mainframe.search.list.buttons[buttonnum].link = link
                        T.F.mainframe.search.list.buttons[buttonnum]:Show()
                        T.F.mainframe.search.enddot:Show()
                        if idx == total_items then
                            T.F.mainframe.search.enddot:Hide()
                        end
                    else
                        B.print("Not found "..tostring(link))
                    end
                end
            end
        end
    end

    function B.ScrollbarUpdateAuctions()

        local time = GetTime()
        local r, g, b, k, v, idx, link, hexcolor, ilvl
        local name, link, quality, iLevel, reqLevel, class, subclass, maxStack, equipSlot, texture, vendorPrice
        local auctions = B.GetPlayerBank('auctions')[1]
        local total_items = #auctions
        local currentscrollbarvalue = FauxScrollFrame_GetOffset(T.F.mainframe.character.auctions.scroll)
        local rows = B.RowsToFit(T.F.mainframe.character.auctions, 28)

        FauxScrollFrame_Update(T.F.mainframe.character.auctions.scroll, total_items, rows, 28)
        for x = 1, rows do
            -- Clear buttons
            B.ResetButton(T.F.mainframe.character.auctions.list.buttons[x])
            --T.F.mainframe.character.auctions.list.buttons[x]:SetBackdropColor(T.THEME.r, T.THEME.g, T.THEME.b, T.THEME.a)
            T.F.mainframe.character.auctions.list.buttons[x]:SetBackdropBorderColor(T.THEME.r, T.THEME.g, T.THEME.b, 1)
            T.F.mainframe.character.auctions.list.buttons[x].leftstring:SetText("")
            T.F.mainframe.character.auctions.list.buttons[x].rightstring.time:SetVertexColor(0.02, 0.76, 0.31, 1)
            --T.F.mainframe.character.auctions.list.buttons[x].rightstring.bg:SetVertexColor(T.THEME.r, T.THEME.g, T.THEME.b, 1)
            T.F.mainframe.character.auctions.list.buttons[x]:Hide()
            T.F.mainframe.character.auctions.enddot:Hide()
        end

        local total = 0
        for x = 1, #auctions do
            total = total + auctions[x].price
        end

        for x = 1, rows do
            -- Get the appropriate item, which will be x + value
            idx = x + currentscrollbarvalue

            if auctions[idx] then
                if strmatch(auctions[idx].link, "|Hbattlepet:") then
                    name, link, quality, iLevel, reqLevel, class, subclass, maxStack, equipSlot, texture, vendorPrice = B.GetPetInfo(auctions[idx].link)
                else
                    name, link, quality, iLevel, reqLevel, class, subclass, maxStack, equipSlot, texture = GetItemInfo(auctions[idx].link)
                end
                if name then
                    r, g, b, hexcolor = GetItemQualityColor(quality)

                    T.F.mainframe.character.auctions.list.buttons[x].quality = quality
                    T.F.mainframe.character.auctions.list.buttons[x].name = name
                    T.F.mainframe.character.auctions.list.buttons[x].icon:SetTexture(texture)
                    T.F.mainframe.character.auctions.list.buttons[x].leftstring:SetText("|c"..hexcolor..name.."|r")
                    T.F.mainframe.character.auctions.list.buttons[x].rightstring:SetText(GetMoneyString(auctions[idx].price, true))
                    if auctions[idx].count and auctions[idx].count > 1 then
                        T.F.mainframe.character.auctions.list.buttons[x].text:SetText(auctions[idx].count)
                    end

                    if (class == L[9004] or class == L[9005]) and (quality and quality > 1) then -- ILVL for green quality or higher
                        ilvl = B.DetectILVL(link)
                        if ilvl then
                            T.F.mainframe.character.auctions.list.buttons[x].ilvl:SetText(ilvl > 1 and ilvl or nil)
                            T.F.mainframe.character.auctions.list.buttons[x].ilvl:SetTextColor(r, g, b, 1)
                        else
                            T.F.mainframe.character.auctions.list.buttons[x].ilvl:SetText(iLevel > 1 and iLevel or nil)
                            T.F.mainframe.character.auctions.list.buttons[x].ilvl:SetTextColor(r, g, b, 1)
                        end
                    end

                    for _, v in ipairs(T.TIMEINDEX) do
                        k = T.TIME[v]
                        if (auctions[idx].time - time) <= v then
                            T.F.mainframe.character.auctions.list.buttons[x].rightstring.time:SetVertexColor(k.r, k.g, k.b, k.a)
                            break
                        end
                    end

                    T.F.mainframe.character.auctions.list.buttons[x].link = auctions[idx].link
                    T.F.mainframe.character.auctions.list.buttons[x]:Show()
                    T.F.mainframe.character.auctions.enddot:Show()
                    if idx == total_items then
                        T.F.mainframe.character.auctions.enddot:Hide()
                    end
                end
            end
        end
        T.F.mainframe.character.auctions.tab:SetNormalTexture(GetCoinIcon(total / 100))
        T.F.mainframe.character.auctions.total = total
        T.F.mainframe.character.auctions.totalitems = total_items
    end

    function B.ScrollbarUpdateMail()

        local time, days = GetTime(), 0
        local r, g, b, k, v, idx, link, hexcolor, ilvl
        local name, link, quality, iLevel, reqLevel, class, subclass, maxStack, equipSlot, texture, vendorPrice
        local mail = B.GetPlayerBank('mail')
        local total_items = #mail[1]
        local currentscrollbarvalue = FauxScrollFrame_GetOffset(T.F.mainframe.character.mail.scroll)
        local rows = B.RowsToFit(T.F.mainframe.character.mail, 28)

        FauxScrollFrame_Update(T.F.mainframe.character.mail.scroll, total_items, rows, 28)
        for x = 1, rows do
            -- Clear buttons
            B.ResetButton(T.F.mainframe.character.mail.list.buttons[x])
            --T.F.mainframe.character.mail.list.buttons[x]:SetBackdropColor(T.THEME.r, T.THEME.g, T.THEME.b, T.THEME.a)
            T.F.mainframe.character.mail.list.buttons[x]:SetBackdropBorderColor(T.THEME.r, T.THEME.g, T.THEME.b, 1)
            T.F.mainframe.character.mail.list.buttons[x].leftstring:SetText("")
            T.F.mainframe.character.mail.list.buttons[x].rightstring.time:SetVertexColor(0.02, 0.76, 0.31, 1)
            --T.F.mainframe.character.mail.list.buttons[x].rightstring.bg:SetVertexColor(T.THEME.r, T.THEME.g, T.THEME.b, 1)
            T.F.mainframe.character.mail.list.buttons[x].rightstring:SetText("")
            T.F.mainframe.character.mail.list.buttons[x].rightstring:SetFont(T.F.mainframe.character.mail.list.buttons[x].rightstring:GetFont(), 12)
            T.F.mainframe.character.mail.list.buttons[x]:Hide()
            T.F.mainframe.character.mail.enddot:Hide()
        end

        for x = 1, rows do
            -- Get the appropriate item, which will be x + value
            local mail = mail[1]
            idx = x + currentscrollbarvalue

            if mail[idx] then
                if strmatch(mail[idx].link, "|Hbattlepet:") then
                    name, link, quality, iLevel, reqLevel, class, subclass, maxStack, equipSlot, texture, vendorPrice = B.GetPetInfo(mail[idx].link)
                else
                    name, link, quality, iLevel, reqLevel, class, subclass, maxStack, equipSlot, texture = GetItemInfo(mail[idx].link)
                end
                
                if name then
                    r, g, b, hexcolor = GetItemQualityColor(quality)

                    T.F.mainframe.character.mail.list.buttons[x].quality = quality
                    T.F.mainframe.character.mail.list.buttons[x].name = name
                    T.F.mainframe.character.mail.list.buttons[x].icon:SetTexture(texture)
                    T.F.mainframe.character.mail.list.buttons[x].leftstring:SetText("|c"..hexcolor..name.."|r")
                    T.F.mainframe.character.mail.list.buttons[x].rightstring.time:Hide()

                    days = (mail[idx].time - time)/24/60/60

                    if mail[idx].money and mail[idx].money < 0 then
                        T.F.mainframe.character.mail.list.buttons[x].rightstring:SetText(B.ASGSC(math.abs(mail[idx].money), false, nil, true, nil, 9))
                        if days < 1 then
                            T.F.mainframe.character.mail.list.buttons[x].rightstring.time:SetVertexColor(0.80, 0.16, 0.10, 1)
                        else
                            T.F.mainframe.character.mail.list.buttons[x].rightstring.time:SetVertexColor(0.87, 0.69, 0.14, 1)
                        end
                        T.F.mainframe.character.mail.list.buttons[x].rightstring.time:Show()
                    else
                        T.F.mainframe.character.mail.list.buttons[x].rightstring:SetFont(T.F.mainframe.character.mail.list.buttons[x].rightstring:GetFont(), 10)
                        T.F.mainframe.character.mail.list.buttons[x].rightstring:SetText(B.GetDays(days))
                    end

                    if mail[idx].count and mail[idx].count > 1 then
                        T.F.mainframe.character.mail.list.buttons[x].text:SetText(mail[idx].count)
                    end

                    if (class == L[9004] or class == L[9005]) and (quality and quality > 1) then -- ILVL for green quality or higher
                        ilvl = B.DetectILVL(link)
                        if ilvl then
                            T.F.mainframe.character.mail.list.buttons[x].ilvl:SetText(ilvl > 1 and ilvl or nil)
                            T.F.mainframe.character.mail.list.buttons[x].ilvl:SetTextColor(r, g, b, 1)
                        else
                            T.F.mainframe.character.mail.list.buttons[x].ilvl:SetText(iLevel > 1 and iLevel or nil)
                            T.F.mainframe.character.mail.list.buttons[x].ilvl:SetTextColor(r, g, b, 1)
                        end
                    end

                    T.F.mainframe.character.mail.list.buttons[x].link = mail[idx].link
                    T.F.mainframe.character.mail.list.buttons[x]:Show()
                    T.F.mainframe.character.mail.enddot:Show()
                    if idx == total_items then
                        T.F.mainframe.character.mail.enddot:Hide()
                    end
                end
            end
        end

        T.F.mainframe.character.mail.tab:SetNormalTexture(B.GetMailIcon(total_items))
        T.F.mainframe.character.mail.totalitems = total_items
        T.F.mainframe.character.mail.money = mail['money'] or 0
        T.F.mainframe.character.mail.cod = mail['cod'] or 0
    end

    function B.BankUpdate()

        local page, total, counter, quality, r, g, b
        local bank = B.GetPlayerBank('bank')

        B.ValidateBankBags(bank)

        counter = 0
        for page = 1, T.NUM_BAG_SLOTS + T.NUM_BANKBAGSLOTS do
            if bank[page] then

                if page > 1 then -- Ignore generic bank
                    B.UpdateBagTab(T.F.mainframe.bank.bags[page - 4].tab, bank[page])
                end
            
                B.UpdatePage(T.F.mainframe.bank, bank[page]['slots'], bank[page], counter)
                counter = counter + bank[page]['slots']
            
            elseif page >= T.NUM_BAG_SLOTS + 1 then
                B.UpdateBagTab(T.F.mainframe.bank.bags[page - 4].tab, {})
            end
        end
    end

    function B.PlayerUpdate()

        local page, total, counter, quality, r, g, b
        local player = B.GetPlayerBank('player')

        B.ValidatePlayerBags(player)

        counter = 0
        for page = 0, #player do

            if page > 0 then -- Ignore generic bag
                B.UpdateBagTab(T.F.mainframe.character.bags[page].tab, player[page])
            end

            if player[page] then

                B.UpdatePage(T.F.mainframe.character, player[page]['slots'], player[page], counter)
                counter = counter + player[page]['slots']
            end
        end
    end

    function B.PlayerSlotUpdate(bag, slot)

        local page, counter, item
        local player = B.GetPlayerBank('player')

        B.ValidatePlayerBags(player)

        counter = 0
        for page = 0, #player do

            if page > 0 then -- Ignore generic bag
                B.UpdateBagTab(T.F.mainframe.character.bags[page].tab, player[page])
            end

            if player[page] then
                if page == bag then
                    
                    B.ResetButton(T.F.mainframe.character.buttons[counter + slot])
                    if player[bag][slot] then

                        item = B.GetItem(player[bag][slot].link)
                        if next(item) then B.UpdateButton(T.F.mainframe.character.buttons[counter + slot], item, player[bag][slot]) end
                    end
                end
                counter = counter + player[page]['slots']
            end
        end
    end

    function B.GenericBankSlotUpdate(slot)
        local bank = B.GetPlayerBank('bank')
        B.UpdateSlot(T.F.mainframe.bank.buttons[slot], bank[1][slot])
    end

    function B.BankSlotUpdate(button, page, slot)
        local bank = B.GetPlayerBank('bank')
        B.UpdateSlot(T.F.mainframe.bank.buttons[button], bank[page][slot])
    end

    function B.PlayerEquipmentUpdate()
        local player = B.GetPlayerBank('player')
        B.UpdatePage(T.F.mainframe.character.equipment, 18, player[-1])
    end

    function B.PlayerEquipmentSlotUpdate(slot)
        local player = B.GetPlayerBank('player')
        B.UpdateSlot(T.F.mainframe.character.equipment.buttons[slot], player[-1][slot])
    end

    function B.ReagentSlotUpdate(slot)
        local reagent = B.GetPlayerBank('reagent')
        B.UpdateSlot(T.F.mainframe.reagent[1].buttons[slot], reagent[1][slot])
    end

    function B.VoidSlotUpdate(page, slot)
        local void = B.GetPlayerBank('void')
        B.UpdateSlot(T.F.mainframe.void[page].buttons[slot], void[page][slot])
    end

    function B.GuildSlotUpdate(page, slot)
        local guild = B.GetPlayerBank('guild')
        B.UpdateSlot(T.F.mainframe.guild[page].buttons[slot], guild[page][slot])       
    end

    function B.ReagentUpdate()
        local page, reagent = _, B.GetPlayerBank('reagent')

        for page = 1, T.REAGENT_PAGES do
            B.UpdatePage(T.F.mainframe.reagent[page], T.NUM_REAGENT_SLOTS, reagent[page])
        end
    end

    function B.VoidUpdate()
        local page, void = _, B.GetPlayerBank('void')

        for page = 1, T.VOID_PAGES do
            B.UpdatePage(T.F.mainframe.void[page], T.NUM_VOID_SLOTS, void[page])
        end
    end

    function B.GuildUpdate()

        local page, guild = _, B.GetPlayerBank('guild')

        for page = 1, T.MAX_GUILDBANK_TABS do
            if guild[page] then

                if guild[page].icon then
                    T.F.mainframe.guild[page].tab:SetNormalTexture(guild[page].icon)
                    T.F.mainframe.guild[page].tab.name = guild[page].name
                    T.F.mainframe.guild[page].tab:Show()
                else
                    T.F.mainframe.guild[page].tab:Hide()
                end

                if guild[page].access then -- Validate user has access to view tab
                    T.F.mainframe.guild[page].tab:Enable()
                    T.F.mainframe.guild[page].tab:GetNormalTexture():SetDesaturated(false)
                    B.UpdatePage(T.F.mainframe.guild[page], T.MAX_GUILDBANK_SLOTS_PER_TAB, guild[page])
                else
                    T.F.mainframe.guild[page].tab:Disable()
                    T.F.mainframe.guild[page].tab:GetNormalTexture():SetDesaturated(true)
                end
            else
                T.F.mainframe.guild[page].tab:Hide()
                B.UpdatePage(T.F.mainframe.guild[page], T.MAX_GUILDBANK_SLOTS_PER_TAB, nil)
            end
         end
    end

    function B.RefreshGuildTab()

        local x, name, icon, isViewable
        local guild = T['guild']
        local tabs = GetNumGuildBankTabs()

        if IsInGuild() then guild['name'] = GetGuildInfo('player') end

        for x = 1, T.MAX_GUILDBANK_TABS do
            if x <= tabs then
                if not guild[x] then guild[x] = B.newtable() end
                name, icon, isViewable = GetGuildBankTabInfo(x)
                guild[x]['icon'] = icon
                guild[x]['name'] = name
                guild[x]['access'] = isViewable

                T.F.mainframe.guild[x].tab:SetNormalTexture(icon)
                T.F.mainframe.guild[x].tab.name = name
                T.F.mainframe.guild[x].tab:Enable()
                T.F.mainframe.guild[x].tab:GetNormalTexture():SetDesaturated(false)
                T.F.mainframe.guild[x].tab:Show()
            else
                guild[x] = B.deltable(guild[x])
                T.F.mainframe.guild[x].tab:Disable()
                T.F.mainframe.guild[x].tab:GetNormalTexture():SetDesaturated(true)
                T.F.mainframe.guild[x].tab:Hide()
            end
        end
    end

    function B.RefreshBankBags()
        local total, page
        local bank = T['bank']

        for bag = T.ITEM_INVENTORY_BANK_BAG_OFFSET + 1, T.ITEM_INVENTORY_BANK_BAG_OFFSET + T.NUM_BANKBAGSLOTS do
            B.RefreshBankBag(bank, bag)

            if bag > 1 then -- Ignore generic bank
                B.UpdateBagTab(T.F.mainframe.bank.bags[bag - 4].tab, bank[bag])
            end
        end

        B.ValidateBankBags(bank)
    end

    function B.RefreshBankBag(bank, bag)

        if not bank[bag] then bank[bag] = B.newtable() end
            
        local inventoryID = BankButtonIDToInvSlotID(bag, 1) - 4
        local baglink = GetInventoryItemLink("player", inventoryID)
        local slots = GetContainerNumSlots(bag)

        bank[bag]['slots'] = slots
        bank[bag]['link'] = baglink

        B.ValidateBankBags(bank)
    end

    function B.RefreshPlayerBag(player, bag)
        local slots

        if not player[bag] then
            player[bag] = B.newtable()
            slots = GetContainerNumSlots(bag)
            player[bag]['slots'] = slots

            if bag > 0 then
                local inventoryID = ContainerIDToInventoryID(bag)
                local baglink = GetInventoryItemLink("player", inventoryID)
                player[bag]['link'] = baglink
            end
        end
    end

    function B.UpdateBagTab(tab, data)
        local quality, r, g, b

        if data.link then
            quality = B.GetQualityFromLink(data.link)
            r, g, b = GetItemQualityColor(quality)

            tab:SetBackdropBorderColor(r, g, b, 1)
            tab.link = data.link
            tab.quality = quality
            tab:SetNormalTexture(GetItemIcon(data.link))
            tab:Show()
        else
            tab:Hide()
        end
    end

    function B.UpdatePage(page, maxslots, data, optstart)
        local x, item
        optstart = optstart or 0

        for x = 1, maxslots do
            B.UpdateSlot(page.buttons[x + optstart], data and data[x] or nil)
        end
    end

    function B.UpdateSlot(button, data)
        local item

        if button then
            B.ResetButton(button)

            if data then
                item = B.GetItem(data.link)

                if next(item) and (button.link ~= data.link or button.count ~= (data.count or 1)) then
                    B.UpdateButton(button, item, data)
                end
            end
        end
    end

    local item
    function B.GetItem(itemlink)

        local itemid, name, link, quality, iLevel, reqLevel, class, subclass, maxStack, equipSlot, texture, vendorPrice
        item = item or B.newtable()

        if T.BW.loaded then
            
            if strmatch(itemlink, "|Hbattlepet:") then
                name, link, quality, iLevel, reqLevel, class, subclass, maxStack, equipSlot, texture, vendorPrice = B.GetPetInfo(itemlink)
            elseif strmatch(itemlink, "|Hkeystone:") then
                name, link, quality, iLevel, reqLevel, class, subclass, maxStack, equipSlot, texture, vendorPrice = B.GetKeystoneInfo(itemlink)
            else
                name, link, quality, iLevel, reqLevel, class, subclass, maxStack, equipSlot, texture, vendorPrice = GetItemInfo(itemlink)
            end

            -- Reuse the same table to avoid generating garbage
            item[1] = name
            item[2] = link
            item[3] = quality
            item[4] = iLevel
            item[5] = reqLevel
            item[6] = class
            item[7] = subclass
            item[8] = maxStack
            item[9] = equipSlot
            item[10] = texture
            item[11] = vendorPrice


            itemid = B.GetItemIdFromLink(itemlink)

            if not next(item) and itemid then -- Item is not cached yet
                T.BW.lookuprequest = true
                T.BW.lookup[tonumber(itemid)] = true
            end
        end

        return item
    end

    function B.UpdateButton(button, item, data)
        local r, g, b, ilvl

        r, g, b = GetItemQualityColor(item[3])

        button.name = item[1]
        button.count = data.count or 1
        button.quality = item[3]
        button.link = item[2]
        button.class = item[6]
        button.subclass = item[7]
        button.equipslot = item[9]
        button.sell = item[11] and item[11] > 0 and item[11] or nil

        button.icon:SetTexture(item[10])
        button.icon:SetTexCoord(0.1, 0.9, 0.1, 0.9)
        
        if (data.count or 1) > 1 then button.text:SetText(data.count) end
        if item[3] > 1 then button:SetBackdropBorderColor(r, g, b, 1) end

        if item[6] == L[9006] then -- QUEST
            button:SetBackdropBorderColor(1, 1, 0, 1)
            button.quest = true

        elseif (item[6] == L[9004] or item[6] == L[9005]) and (button.quality and button.quality > 1) then -- ILVL for green quality or higher
            ilvl = T.BW.ilvlloaded and B.DetectILVL(item[2]) or item[4] > 1 and item[4] or nil

            button.ilvl:SetText(ilvl)
            button.ilvl:SetTextColor(r, g, b, 1)
        end
    end

    function B.ResetButton(button)

        button.icon:SetTexture("")
        button:SetBackdropBorderColor(T.BORDER_SLOT.r, T.BORDER_SLOT.g, T.BORDER_SLOT.b, T.BORDER_SLOT.a)
        button.link = nil
        button.quality = nil
        button.quest = nil
        button.name = nil
        button.count = nil
        button.text:SetText("")
        button.ilvl:SetText("")
    end

--[[//////////////////////////////////////////////////

    SEARCH FUNCTIONS

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]

    function B.Search(text)

        if T.BW.loaded then
            if text == "" or text == L[9007] then text = false end

            T.F.mainframe.search.scroll.ScrollBar:SetValue(0)

            B.cleantable(T.BW.items)

            B.BankSearch(text)
            B.ReagentSearch(text)
            B.VoidSearch(text)
            B.GuildSearch(text)
            B.PlayerSearch(text)
            B.PlayerEquipmentSearch(text)
            B.MailSearch(text)
            B.AuctionSearch(text)
            B.SearchAlts(text)

            B.ScrollbarUpdate()

            if text then
                T.F.mainframe.countbutton.labelcount:SetText("|cffffffff"..T.F.mainframe.countbutton.count.."|r")
                T.F.mainframe.countbutton:Show()
                if not T.F.mainframe.search:IsVisible() and not T.F.mainframe.lastbank then
                    T.F.mainframe.countbutton:Click()
                end
            else
                T.F.mainframe.countbutton:Hide()
            end
        end
    end

    function B.BankSearch(text)

        local x, page, counter, button, buttons, name
        local bank = B.GetPlayerBank('bank')

        local counter = 1
        for page = 1, #bank do
            if bank[page] then

                if page > 1 then -- ignore generic bank
                    local tab = page - 4

                    if text then T.F.mainframe.bank.bags[tab].tab.search:Show()
                    else T.F.mainframe.bank.bags[tab].tab.search:Hide() end
                end

                buttons = T.F.mainframe.bank.buttons

                for x = 1, bank[page]['slots'] do
                    button = buttons[counter]

                    if button.name and text then
                        if B.find(text, button) then
                            button.loc = "bank."..(page == 1 and page or page - 4)
                            B.SearchSlot(button)
                            if page > 1 then local tab = page - 4; T.F.mainframe.bank.bags[tab].tab.search:Hide() end
                        else
                            button.search:Show()
                        end
                    else
                        button.search:Hide()
                    end
                    counter = counter + 1
                end
            end
        end
    end

    function B.PlayerSearch(text)

        local x, page, counter, button, buttons, name
        local player = B.GetPlayerBank('player')

        local counter = 1
        for page = 0, #player do
            if player[page] then

                if page > 0 then -- ignore generic bag
                    local tab = page

                    if text then T.F.mainframe.character.bags[tab].tab.search:Show()
                    else T.F.mainframe.character.bags[tab].tab.search:Hide() end
                end

                buttons = T.F.mainframe.character.buttons
                for x = 1, player[page]['slots'] do
                    button = buttons[counter]

                    if button.name and text then
                        if B.find(text, button) then
                            button.loc = "player."..(page == 0 and 1 or page + 1)
                            B.SearchSlot(button)
                            if page > 0 then local tab = page; T.F.mainframe.character.bags[tab].tab.search:Hide() end
                        else
                            button.search:Show()
                        end
                    else
                        button.search:Hide()
                    end
                    counter = counter + 1
                end
            end
        end
    end

    function B.PlayerArtifactSearch()

        local x, page, counter, button, buttons, name
        local player = T['player']

        local counter = 1
        for page = 0, #player do
            if player[page] then

                if page > 0 then -- ignore generic bag
                    T.F.mainframe.character.bags[page].tab.search:Show()
                end

                buttons = T.F.mainframe.character.buttons
                for x = 1, player[page]['slots'] do
                    button = buttons[counter]

                    if button.name then
                        if IsArtifactPowerItem(B.GetItemIdFromLink(button.link)) then
                            if not T.F.mainframe.character.artifact.total then T.F.mainframe.character.artifact.total = B.DetectAP(B.GetBag(button)) end
                            if page > 0 then T.F.mainframe.character.bags[page].tab.search:Hide() end
                        else
                            button.search:Show()
                        end
                    else
                        button.search:Hide()
                    end
                    counter = counter + 1
                end
            end
        end
    end

    function B.PlayerEquipmentSearch(text)

        local x, button, buttons, name
        local player = B.GetPlayerBank('player')

        if not player[-1] then return end

        buttons = T.F.mainframe.character.equipment.buttons
        for x = 1, T.EQUIPMENT_SLOTS do
            button = buttons[x]

            if button.name and text then
                if B.find(text, button) then
                    button.loc = "equip.1"
                    B.SearchSlot(button)
                else
                    button.search:Show()
                end
            else
                button.search:Hide()
            end

        end
    end

    function B.BankBagSearch(bag)

        local bank = B.GetPlayerBank('bank')
        local x, page, counter, button, buttons

        local counter = 1
        for page = 1, #bank do
            if bank[page] then

                if not bag or page == bag then
                    for x = 1, bank[page]['slots'] do
                        T.F.mainframe.bank.buttons[counter].search:Hide()
                        counter = counter + 1
                    end
                else
                    for x = 1, bank[page]['slots'] do
                        T.F.mainframe.bank.buttons[counter].search:Show()
                        counter = counter + 1
                    end
                end
            end
        end
    end

    function B.PlayerBagSearch(bag)

        local player = B.GetPlayerBank('player')
        local x, page, counter, button, buttons

        local counter = 1
        for page = 0, #player do
            if player[page] then

                if not bag or page == bag then
                    for x = 1, player[page]['slots'] do
                        T.F.mainframe.character.buttons[counter].search:Hide()
                        counter = counter + 1
                    end
                else
                    for x = 1, player[page]['slots'] do
                        T.F.mainframe.character.buttons[counter].search:Show()
                        counter = counter + 1
                    end
                end
            end
        end
    end

    function B.ReagentSearch(text)

        local x, page, button, buttons
        local reagent = B.GetPlayerBank('reagent')

        for page = 1, T.REAGENT_PAGES do
            buttons = T.F.mainframe.reagent[page].buttons
            for x = 1, T.NUM_REAGENT_SLOTS do
                button = buttons[x]

                if button.name and text then
                    if B.find(text, button) then
                        button.loc = "reagent."..page
                        B.SearchSlot(button)
                    else
                        button.search:Show()
                    end
                else
                    button.search:Hide()
                end
            end
        end
    end

    function B.VoidSearch(text)

        local x, page, button, buttons
        local void = B.GetPlayerBank('void')

        for page = 1, T.VOID_PAGES do

            if text then T.F.mainframe.void[page].tab.search:Show() 
            else T.F.mainframe.void[page].tab.search:Hide() end

            buttons = T.F.mainframe.void[page].buttons

            for x = 1, T.NUM_VOID_SLOTS do
                button = buttons[x]

                if button.name and text then
                    if B.find(text, button) then
                        button.loc = "void."..page
                        B.SearchSlot(button)
                        T.F.mainframe.void[page].tab.search:Hide()
                    else
                        button.search:Show()
                    end
                else
                    button.search:Hide()
                end
            end
        end
    end

    function B.GuildSearch(text)

        local x, page, button, buttons
        local guild = B.GetPlayerBank('guild')

        for page = 1, T.MAX_GUILDBANK_TABS do

            if text then T.F.mainframe.guild[page].tab.search:Show()
            else T.F.mainframe.guild[page].tab.search:Hide() end

            if guild[page] and guild[page].access then

                buttons = T.F.mainframe.guild[page].buttons

                for x = 1, T.MAX_GUILDBANK_SLOTS_PER_TAB do
                    button = buttons[x]

                    if button.name and text then
                        if B.find(text, button) then
                            button.loc = "guild."..page
                            B.SearchSlot(button)
                            T.F.mainframe.guild[page].tab.search:Hide()
                        else
                            button.search:Show()
                        end
                    else
                        button.search:Hide()
                    end
                end
            end
        end
    end

    function B.MailSearch(text)
        local item, link, info, temp, result
        local mail = B.GetPlayerBank('mail')[1]

        if not text then return end

        if next(mail) then
            for x = 1, #mail do

                link = mail[x]['link']
                info = strmatch(link, "|Hbattlepet:") and {B.GetPetInfo(link)} or {GetItemInfo(link)}

                temp = B.newtable()
                temp.name = info[1]
                temp.quality = info[3]
                temp.link = link
                temp.class = info[6]
                temp.subclass = info[7]
                temp.equipslot = info[9]

                if B.find(text, temp) then
                    result = T.BW.items("mail", 1, link) -- use metatable __call to group items
                    result['count'] = result['count'] + (mail[x]['count'] or 1)
                    T.F.mainframe.countbutton.count = T.F.mainframe.countbutton.count + (mail[x]['count'] or 1)
                end
            end
        end
    end

    function B.AuctionSearch(text)
        local item, link, info, temp, result
        local auctions = B.GetPlayerBank('auctions')[1]

        if not text then return end

        if next(auctions) then
            for x = 1, #auctions do

                link = auctions[x]['link']
                info = strmatch(link, "|Hbattlepet:") and {B.GetPetInfo(link)} or {GetItemInfo(link)}

                temp = B.newtable()
                temp.name = info[1]
                temp.quality = info[3]
                temp.link = link
                temp.class = info[6]
                temp.subclass = info[7]
                temp.equipslot = info[9]

                if B.find(text, temp) then
                    result = T.BW.items("auctions", 1, link) -- use metatable __call to group items
                    result['count'] = result['count'] + (auctions[x]['count'] or 1)
                end
            end
        end
    end

    function B.SearchAlts(text)
        local k, v, p, b, player, faction

        if not text then return end

        if BWsavedtable['enablesearchallservers'] then
            for k, v in pairs(BWsavedtable) do
                if not T.OPTIONS[k] then
                    for p, b in pairs(v) do
                        B.SearchServer(k, p, b, text)
                    end
                end
            end
        else -- Only same server and faction as selected player
            for p, b in pairs(BWsavedtable[T.ACTIVE_TABLE.server]) do
                B.SearchServer(T.ACTIVE_TABLE.server, p, b, text)
            end
        end
    end

    local search_item
    function B.SearchServer(server, playerfaction, t, text)
        local player, faction, bank, bags, item, link, info, temp

        search_item = search_item or B.newtable()
        player, faction = playerfaction:match"([^.]*).(.*)"

        if player == T.ACTIVE_TABLE.player and faction == string.sub(T.ACTIVE_TABLE.faction, 1, 1) then return end -- skip loaded character

        if BWsavedtable['enablesearchallfactions'] or faction == string.sub(T.ACTIVE_TABLE.faction, 1, 1) then
            
            for _, bank in pairs(t) do-- banks, bags
                for _, bags in pairs(bank) do -- bag index, bags
                    if type(bags) == "table" then
                        for _, item in pairs(bags) do
                            if type(item) == "table" then

                                link = item['link']
                                info = B.GetItem(link)

                                if next(info) then

                                    search_item['name'] = info[1]
                                    search_item['quality'] = info[3]
                                    search_item['link'] = link
                                    search_item['class'] = info[6]
                                    search_item['subclass'] = info[7]
                                    search_item['equipslot'] = info[9]

                                    if B.find(text, search_item) then
                                        B.SearchAltSlot(server.."."..playerfaction, link, item['count'] or 1)
                                    end
                                end
                            end
                        end
                    end
                end
            end
        end
    end

    function B.SearchSlot(button)

        button.search:Hide()

        local loc, num = button.loc:match"([^.]*).(.*)"
        num = tonumber(num)

        local result = T.BW.items(loc, num, button.link) -- use metatable __call to group items
        result['count'] = result['count'] + button.count
        T.F.mainframe.countbutton.count = T.F.mainframe.countbutton.count + button.count
    end

    function B.SearchAltSlot(alt, link, count)

        local result = T.BW.items("alt", alt, link) -- use metatable __call to group items
        result['count'] = result['count'] + count
    end

--[[//////////////////////////////////////////////////

    TOOLTIP CACHE FUNCTIONS

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]

    function B.CacheItems(cache, itemid, count)
        if not itemid then return end -- keystone don't have itemids

        itemid = tonumber(itemid)

        if not cache[itemid] then
            cache[itemid] = 0
        end

        cache[itemid] = cache[itemid] + count
    end

    function B._CacheItems(cache, location, bank)
        local bags, item, b

        if not bank then return end

        if not cache[location] then
            cache[location] = B.newtable()
        end

        table.wipe(cache[location])

        if location == "equipped" then
            for _, item in pairs(bank) do
                if type(item) == "table" then

                    B.CacheItems(cache[location], B.GetItemIdFromLink(item['link']), item['count'] or 1)
                end
            end
        else
            for b, bags in pairs(bank) do
                if (location == "player" and b ~= "class" and b > -1) or location ~= "player" then
                    if type(bags) == "table" then
                        for _, item in pairs(bags) do
                            if type(item) == "table" then

                                B.CacheItems(cache[location], B.GetItemIdFromLink(item['link']), item['count'] or 1)
                            end
                        end
                    end
                end
            end
        end
    end

    function B.CacheMainItems(bank)
        -- Cache everything from active player into a workable table by itemid - count only.
        local banks, selected
        local player = UnitName('player')
        local faction = string.sub(UnitFactionGroup('player'), 1, 1)
        local server = GetRealmName()
        local cachename = player.."."..faction

        T.BW.tooltipcache[server] = T.BW.tooltipcache[server] or B.newtable()
        T.BW.tooltipcache[server][cachename] = T.BW.tooltipcache[server][cachename] or B.newtable()

        if not bank then -- Cache everything
            banks = {"bank", "reagent", "void", "guild", "player", "equipped", "auctions", "mail"}
            for bank = 1, #banks do
                selected = banks[bank] == "equipped" and T['player'][-1] or T[banks[bank]]
                B._CacheItems(T.BW.tooltipcache[server][cachename], banks[bank], selected)
            end
        else
            selected = bank == "equipped" and T['player'][-1] or T[bank]
            B._CacheItems(T.BW.tooltipcache[server][cachename], bank, selected)
        end
    end

    function B.CacheAltItems() -- Run only once
        local player, faction, p, b, banks, bank, selected

        for k, v in pairs(BWsavedtable) do
            if not T.OPTIONS[k] and type(v) == "table" then
                for p, b in pairs(v) do
                    B.CacheAlt(k, p, b)
                end
            end
        end
    end

    function B.CacheAlt(server, playerfaction, t)
        local player, faction, cachename, selected

        player, faction = playerfaction:match"([^.]*).(.*)"

        if player == UnitName('player') and faction == string.sub(UnitFactionGroup('player'), 1, 1) then return end

        if BWsavedtable['enablesearchallfactions'] or faction == string.sub(UnitFactionGroup('player'), 1, 1) then
            cachename = player.."."..faction

            T.BW.tooltipcache[server] = T.BW.tooltipcache[server] or B.newtable()
            T.BW.tooltipcache[server][cachename] = B.newtable()

            local banks = {"bank", "reagent", "void", "guild", "player", "equipped", "auctions", "mail"}
            for bank = 1, #banks do
                selected = banks[bank] == "equipped" and t['player'][-1] or t[banks[bank]]
                B._CacheItems(T.BW.tooltipcache[server][cachename], banks[bank], selected)
            end
        end
    end

--[[//////////////////////////////////////////////////

    CLEAN UP CREW

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]

    function B.CleanAuctions(bank)

        local x
        local auctions = T['auctions']
        local time = GetTime()

        for x = #auctions[1], 1, -1 do
            if auctions[1][x] then
                if auctions[1][x]['time'] <= time then -- expired
                    table.remove(auctions[1][x], x)
                end
            end
        end
    end

    function B.CleanAltsAuctions()

        local k, v, p, b

        for k, v in pairs(BWsavedtable) do
            if not T.OPTIONS[k] and type(v) == "table" then
                for p, b in pairs(v) do
                    if p ~= UnitName('player').."."..string.sub(UnitFactionGroup('player'), 1, 1) then
                        B.CleanAuctions(b['auctions'])
                    end
                end
            end
        end
    end

    function B.CleanMail(bank)

        local x
        local mail = bank or T['mail']
        local time = GetTime()

        for x = #mail[1], 1, -1 do
            if mail[1][x] then
                if mail[1][x]['time'] <= time then -- expired
                    table.remove(mail[1][x], x)
                end
            end
        end
    end

    function B.CleanAltsMail()

        local k, v, p, b

        for k, v in pairs(BWsavedtable) do
            if not T.OPTIONS[k] and type(v) == "table" then
                for p, b in pairs(v) do
                    if p ~= UnitName('player').."."..string.sub(UnitFactionGroup('player'), 1, 1) then
                        B.CleanMail(b['mail'])
                    end
                end
            end
        end
    end
