-- GLOBALS: unpack, select
-- GLOBALS: BanksWatch, ReagentBankFrame, StackSplitFrame, SlashCmdList, DEFAULT_CHAT_FRAME, GetAddOnMetadata
-- GLOBALS: SLASH_BanksWatch1, SLASH_BanksWatch2, SLASH_BanksWatch3, SLASH_BanksWatch4, SLASH_BanksWatch5, SLASH_BanksWatch6, SLASH_BanksWatch7, SLASH_BanksWatch8

--[[//////////////////////////////////////////////////

    GLO-LO

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]

    local table = table
    local next = next
    local tonumber = tonumber
    local pairs = pairs

    local UISpecialFrames = UISpecialFrames
    local MoneyInputFrame_GetCopper = MoneyInputFrame_GetCopper
    local GetMoneyString = GetMoneyString
    local ReagentBankFrame_OnShow = ReagentBankFrame_OnShow
    local IsBagOpen, OpenBag, CloseBag = IsBagOpen, OpenBag, CloseBag

    local B, L, T, E = unpack(select(2, ...))
    local events = {}
    local bag_to_update = nil
    local bag_update = {}
    local local_elapsed = 0

    BanksWatch = E

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]

    local function setminimum_show(self)
        if T.CONFIG['minimumsoundtracker'] then
            self.text:SetText(self.text:GetText().."\n"..GetMoneyString(T.CONFIG['minimumsoundtracker']))
        end
    end

    local function setminimum_accept(self)
        T.CONFIG['minimumsoundtracker'] = tonumber(MoneyInputFrame_GetCopper(self.moneyInputFrame))
    end

    local function settreasure_show(self)
        if T.CONFIG['treasuresoundtracker'] then
            self.text:SetText(self.text:GetText().."\n"..GetMoneyString(T.CONFIG['treasuresoundtracker']))
        end
    end

    local function settreasure_accept(self)
        T.CONFIG['treasuresoundtracker'] = tonumber(MoneyInputFrame_GetCopper(self.moneyInputFrame))
    end

--[[//////////////////////////////////////////////////

    FUNCTIONS TRIGGERED VIA XML bankswatch.xml

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]

    local function bank_opened(bank)

        local banks = {
            ['bank'] = T.F.mainframe.bank,
            ['void'] = T.F.mainframe.void,
            ['guild'] = T.F.mainframe.guild
        }

        T.F.mainframe.characterbutton:SetBackdropBorderColor(0.01, 0.79, 0.15, 1)
        T.F.mainframe.character:SetClampedToScreen(true)
        T.F.mainframe.characterbutton:SetScript("OnClick", function(self, button)

            if not T.F.mainframe.lastbank then 
                B.SetFrameLocation(true) -- force refresh to reclip everything together
            else
                B.AnchorBags(banks[bank])
            end

            if T.F.mainframe.character:IsVisible() then
                T.F.mainframe.character:Hide()
            else
                T.F.mainframe.character:Show()
            end
        end)

        if T.CONFIG['openbank'] and not banks[bank]:IsVisible() then T.F.mainframe.character:Hide(); BanksWatch.ToggleBank(bank) end
        if T.CONFIG['openplayer'] and not T.F.mainframe.character:IsVisible() then T.F.mainframe.characterbutton:Click() end
    end

    local function bank_closed(bank)

        B.HideAll()
        T.F.mainframe.lastbank = nil
        T.F.mainframe.characterbutton:SetBackdropBorderColor(0, 0, 0, 0)
        T.F.mainframe.characterbutton:SetScript("OnClick", B.I.Character.Click)
        T.F.mainframe.character:SetClampedToScreen(false)
        B.SetFrameLocation(true) -- force refresh to reclip everything together
    end

    local function genericbank_update(slot)

        if slot <= T.NUM_BANKGENERIC_SLOTS then -- Generic bank modified
            B.GenericBankSlotScan(slot)

        elseif slot > T.NUM_BANKGENERIC_SLOTS then -- Bank bag modified, will trigger BAG_UPDATE
            bag_to_update = slot - T.NUM_BANKGENERIC_SLOTS + 4
        end
    end

    local function guildbank_update()

        T.ISGUILDOPEN = true
        B.RefreshGuildTab()
        bank_opened('guild')
    end

    local function tooltip_update(self, elapsed)
        if not elapsed then return end

        local_elapsed = local_elapsed + elapsed
        if local_elapsed > 5 then
            local_elapsed = 0
            T.BW.ilvlloaded = true
            B.UpdateBanks()
            self:SetScript("OnUpdate", nil)
        end
    end

    local function bagdelay_update(self, elapsed)

        if not elapsed then return end -- Otherwise it will infinite loop

        if T.ISBANKOPEN and T.BW.forceBankScan and next(bag_update) then
            local k, v
            for k, v in pairs(bag_update) do -- try to minimize how many updates to bank bags
                B.BankBagScan(k)
                bag_update[k] = nil
            end
            B.CacheMainItems('bank')
            BanksWatchframe:SetScript("OnUpdate", nil)
        end
    end

    function BanksWatch:OnLoad()

        self:RegisterEvent('BANKFRAME_OPENED')
        self:RegisterEvent('BANKFRAME_CLOSED')
        self:RegisterEvent('PLAYERBANKSLOTS_CHANGED')
        self:RegisterEvent('PLAYERREAGENTBANKSLOTS_CHANGED')
        self:RegisterEvent('VOID_STORAGE_OPEN')
        self:RegisterEvent('VOID_STORAGE_CLOSE')
        self:RegisterEvent('GUILDBANKFRAME_OPENED')
        self:RegisterEvent('GUILDBANK_UPDATE_TABS')
        self:RegisterEvent('PLAYER_ENTERING_WORLD')
        self:RegisterEvent('VARIABLES_LOADED')
        self:RegisterEvent('PLAYER_REGEN_DISABLED')
        self:RegisterEvent('GET_ITEM_INFO_RECEIVED')
        self:RegisterEvent('PLAYER_EQUIPMENT_CHANGED')
        self:RegisterEvent('PLAYERBANKBAGSLOTS_CHANGED')
        self:RegisterEvent('AUCTION_OWNED_LIST_UPDATE')
        self:RegisterEvent('LOADING_SCREEN_DISABLED')
        self:RegisterEvent('MAIL_INBOX_UPDATE')
        self:RegisterEvent('PLAYER_MONEY')

        StackSplitFrame:SetFrameStrata('DIALOG') -- For some reason, this blizzard frame is not set as a dialog

        SLASH_BanksWatch1 = "/BW"
        SLASH_BanksWatch2 = "/bw"
        SLASH_BanksWatch3 = "/Bw"
        SLASH_BanksWatch4 = "/bW"
        SLASH_BanksWatch5 = "/Bankswatch"
        SLASH_BanksWatch6 = "/BanksWatch"
        SLASH_BanksWatch7 = "/BANKSWATCH"
        SLASH_BanksWatch8 = "/bankswatch"
        SlashCmdList['BanksWatch'] = B.Main

        DEFAULT_CHAT_FRAME:AddMessage(T.MSGC.DEFAULT..L[10000])

        B.U.StaticPopupDialogs("BWsetminimum", L[10053], nil, nil, true, setminimum_show, setminimum_accept)
        B.U.StaticPopupDialogs("BWsettreasure", L[10054], nil, nil, true, settreasure_show, settreasure_accept)

        B.CreateFrames()
        B.M.Create() -- Create Money frame

        table.insert(UISpecialFrames, T.F.mainframe.bank:GetName())
        table.insert(UISpecialFrames, T.F.mainframe.reagent:GetName())
        table.insert(UISpecialFrames, T.F.mainframe.void:GetName())
        table.insert(UISpecialFrames, T.F.mainframe.guild:GetName())
        table.insert(UISpecialFrames, T.F.mainframe.search:GetName())
        table.insert(UISpecialFrames, T.F.mainframe.character:GetName())
    end

    function BanksWatch:OnEvent(event, arg1)
        --B.print(event)
        events[event](self, arg1)
    end

    function events:VARIABLES_LOADED()
        B.print(T.MSGC.INFO.."Running version: "..GetAddOnMetadata("BanksWatch", "Version"), 1)
        B.print(T.MSGC.INFO.."And now my watch begins...", 1)

        B.C.LoadSettings()
        B.CreateLayouts()
    end

    function events:GET_ITEM_INFO_RECEIVED(itemid)

        if T.BW.lookup[itemid] then T.BW.lookup[itemid] = nil end

        if T.BW.lookuprequest and not next(T.BW.lookup) then -- Try to minimize the amount of updates
            T.BW.lookuprequest = false
            B.UpdateBanks()
        end
    end

    function events:PLAYER_ENTERING_WORLD()

        self:RegisterEvent('PLAYER_LEAVING_WORLD')

        if not T.BW.loaded then -- First time loading character
            self:RegisterEvent('PLAYER_REGEN_DISABLED')
            self:RegisterEvent('MERCHANT_SHOW')

            T.BW.items = B.newsearchtable()

            B.Initialize() -- Load or create character
            B.C.Create() -- Config

            if T.CONFIG['minimize'] then T.F.mainframe.togglebutton:Show() end
            if not BWsavedtable['minimapicon']['hide'] and BWconfig['state'] == 0 then T.F.toggleframe:Hide() end

            B.SetFrameLocation()
            B.AdjustTransparency()

            B.M.Frame.Scale(T.CONFIG['scalemoneytracker'] or 125)

            B.CleanAuctions()
            B.CleanMail()
            B.CacheMainItems()

            -- Runs only once
            B.CleanAltsAuctions()
            B.CleanAltsMail()
            B.CacheAltItems()
        end
    end

    function events:PLAYER_LEAVING_WORLD()

        self:UnregisterEvent('PLAYER_LEAVING_WORLD')
        self:UnregisterEvent('BAG_UPDATE')
    end

    function events:LOADING_SCREEN_DISABLED() -- Register BAG_UPDATE here to avoid spam BAG_UPDATE when zoning
        self:RegisterEvent('BAG_UPDATE')
        if not T.BW.loaded then
            T.BW.loaded = true

            B.PlayerBagsScan()
            B.PlayerEquipmentScan()

            self:SetScript("OnUpdate", tooltip_update)
        end
    end

    function events:PLAYERBANKSLOTS_CHANGED(slot)
        if slot then genericbank_update(slot) end
    end

    function events:PLAYERBANKBAGSLOTS_CHANGED(slot)
        if slot then genericbank_update(slot) end
    end

    function events:PLAYER_EQUIPMENT_CHANGED(slot)
        B.PlayerEquipmentSlotScan(slot)
    end

    function events:BAG_UPDATE(bag)

        if bag == 0 or bag <= T.NUM_BAG_SLOTS then
            -- Only run when bags are not open since magic doesn't happen unless bags are open.
            if not IsBagOpen(bag) then B.PlayerBagScan(bag) end

        elseif bag_to_update == bag then
            bag_to_update = nil
            B.RefreshBankBags()
            B.BankBagScan(bag)

        elseif T.ISBANKOPEN and T.BW.forceBankScan and (bag > T.NUM_BAG_SLOTS and bag <= T.NUM_BAG_SLOTS + T.NUM_BANKBAGSLOTS) then
            bag_update[tonumber(bag)] = true
            if not BanksWatchframe:GetScript("OnUpdate") then BanksWatchframe:SetScript("OnUpdate", bagdelay_update) end
        end
    end

    function events:VOID_STORAGE_OPEN()
        
        T.ISVOIDOPEN = true
        bank_opened('void')
    end

    function events:VOID_STORAGE_CLOSE()
        
        T.ISVOIDOPEN = false
        bank_closed('void')
    end

    function events:BANKFRAME_OPENED()
        
        T.ISBANKOPEN = true
        ReagentBankFrame_OnShow(ReagentBankFrame) -- force generate to activate the listener
        B.RefreshBankBags()

        if T.BW.forceBankScan then
            B.BankScan()
        else
            local id
            for id = T.NUM_BAG_SLOTS + 1, T.NUM_BAG_SLOTS + T.NUM_BANKBAGSLOTS do -- open all bank bags
                if not IsBagOpen(id) then
                    OpenBag(id)
                    CloseBag(id)
                end
            end
        end

        bank_opened('bank')
    end

    function events:BANKFRAME_CLOSED()

        T.ISBANKOPEN = false
        bank_closed('bank')
    end

    function events:GUILDBANKFRAME_OPENED()

        self:RegisterEvent('GUILDBANKFRAME_CLOSED')
        guildbank_update()
    end

    function events:GUILDBANK_UPDATE_TABS()
        guildbank_update()
    end

    function events:GUILDBANKFRAME_CLOSED()

        self:UnregisterEvent('GUILDBANKFRAME_CLOSED')
        T.ISGUILDOPEN = false
        bank_closed('guild')
    end

    function events:PLAYER_REGEN_DISABLED() -- Entering combat, hide any open bags
        T.F.mainframe.togglebutton:Show()
    end

    function events:PLAYERREAGENTBANKSLOTS_CHANGED(slot)
        B.ReagentSlotScan(slot)
    end

    function events:AUCTION_OWNED_LIST_UPDATE()

        B.AuctionScan()
        B.CacheMainItems('auctions')
        B.CacheMainItems('player')
        B.ScrollbarUpdateAuctions()
    end

    function events:MERCHANT_SHOW()

        T.ISMERCHANTOPEN = true
        self:RegisterEvent('MERCHANT_CLOSED')
    end

    function events:MERCHANT_CLOSED()

        T.ISMERCHANTOPEN = false
        self:UnregisterEvent('MERCHANT_CLOSED')
    end

    function events:MAIL_INBOX_UPDATE()

        B.MailScan()
        B.CacheMainItems('mail')
        B.CacheMainItems('player')
        B.ScrollbarUpdateMail()
    end

    function events:PLAYER_MONEY()
        B.M.Trigger()
    end

--[[//////////////////////////////////////////////////

    FUNCTIONS TRIGGERED VIA XML Bindings.xml

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]

    function BanksWatch:ToggleFrame()

        if not BWsavedtable['minimapicon']['hide'] then
            B.Main("toggle")
            return
        end

        if T.F.mainframe:IsVisible() then
            T.F.mainframe.togglebutton:Show()
            T.F.toggleframe:Show()
        elseif T.F.mainframe.togglebutton:IsVisible() then
            T.F.toggleframe:Hide()
        else
            T.F.toggleframe:Show()
            T.F.mainframe.togglebutton:Hide()
        end
    end

    function BanksWatch:ToggleBank(bank)
        local banks = {
            ['bank'] = T.F.mainframe.bankbutton,
            ['reagent'] = T.F.mainframe.reagentbutton,
            ['character'] = T.F.mainframe.characterbutton,
            ['void'] = T.F.mainframe.voidbutton,
            ['guild'] = T.F.mainframe.guildbutton
        }
        T.F.toggleframe:Show()
        T.F.mainframe.togglebutton:Hide()
        T.F.mainframe:Show()
        banks[bank]:Click()
    end

    function BanksWatch:ToggleMoneyFrame()
        B.Main("money toggle")
    end
